/**
 *
 */


// System Import Headers.
#include <cstdio>
#include <cstdlib>
#include <cctype>
#include <filesystem>
#include <fstream>
#include <iostream>
#include <iterator>
#include <string>
#include <vector>

// User Import Headers.
#ifndef user_headers
    #define user_headers
    #include "structs.h"
    #include "image.h"
#endif


// Method Declaration.
path_validator_struct* validate_path(std::string path_str);
void process_dir(std::string path_str);
void process_file(std::string path_str);
// void validate_path(char* path_str);
// void compute(int a, int b);
// __global__ void cuda_add(int a, int b, int* c);


// Global Variables.

/**
 * Class that holds data for a single image.
 */
class Image {

    private:
        /**
         * Struct to act as a "container" of sorts for each individual rgb pixel.
         * An array of these creates our entire image.
         *
         * According to https://stackoverflow.com/a/36917400, this is basically treated as a sub-class.
         * This allows creating multiple constructors, which will handle different scenarios of pixel input.
         */
        struct Rgb {
            // Struct variables.
            float r, g, b;

            /**
             * Struct constructors.
             *
             * First one acts as a "default" that provides black pixels.
             * When a single float is provided, the second constructor sets all three rgb channels to that value.
             * Finally, when three floats are provided, the last constructor sets rgb channels accordingly.
             */
            Rgb(): r(0), g(0), b(0) {}
            Rgb(float c): r(c), g(c), b(c) {}
            Rgb(float _r, float _g, float _b) : r(_r), g(_g), b(_b) {}

        };

        // Private Variables.
        std::string input_path;
        std::string output_path;
        unsigned int width;
        unsigned int height;
        int bit_depth;
        Rgb *pixel_arr;     // Array of pixel data.


    public:
        // Public Variables.
        static const Rgb pixelBlack, pixelWhite, pixelRed, pixelGreen, pixelBlue, pixelYellow, pixelPink, pixelLiBlue;


        //region Constructors.

        /**
         * Base constructor.
         */
        Image() {
            // Currently test logic.
            std::cout << "\nConstructing Image class." << std::endl;

            // Essentially creates a 10x10 pixel test image.
            input_path = "../input/pbm/400x300/hallway_00.pbm";
            output_path = "../output/test_00.pbm";
            width = 10;
            height = 10;
            bit_depth = 255;
            Rgb c = pixelBlack; // Default color. Can select based on below preset colors.

            // Create and populate image.
            pixel_arr = new Rgb[width * height];
            for (int index = 0; index < (width * height); index++) {
                pixel_arr[index] = c;
            }

            std::cout << "Image class populated." << std::endl;
        }

        /**
         * Constructor that takes input path to image file.
         * Output path is determined based on parsing input path.
         */
        Image(char* user_path) {
            std::cout << "\nConstructing Image class." << std::endl;

            // populate here

            std::cout << "Image class populated." << std::endl;
        }

        //endregion Constructors.


        //region Public Methods.

        /**
         * Displays general properties of image held in class.
         */
        void display_properties() {
            std::cout << std::endl;
            std::cout << "input file path: " << input_path << std::endl;
            std::cout << "output file path: " << output_path << std::endl;
            std::cout << "width: " << width << std::endl;
            std::cout << "height: " << height << std::endl;
            std::cout << "bit depth: " << bit_depth << std::endl;
            std::cout << std::endl;
        }


        /**
         * Saves class image data to output file.
         */
        void save() {
            std::cout << "Attempting to save image to \"" << output_path << "\"." << std::endl;

            // Open output file stream.
            std::ofstream output_file;

            // Attempt to save file data.
            try {
                // Attempt to open file.
                output_file.open(output_path, std::ios::binary | std::ios::out);

                // Check if opened successfully.
                if (output_file.fail()) {
                    // Failed to open.
                    std::cout << "Failed to open file \"" + output_path + "\".\n";
                } else {
                    // Opened successfully.

                    // Write image "header" data.
                    output_file << "P6\n" << width << " " << height << "\n255\n";

                    // Write image pixel data.
                    unsigned char r, g, b;
                    for (int index = 0; index < (width * height); index++) {
                        r = static_cast<float>(std::min(1.f, pixel_arr[index].r) * 255);
                        g = static_cast<float>(std::min(1.f, pixel_arr[index].g) * 255);
                        b = static_cast<float>(std::min(1.f, pixel_arr[index].b) * 255);
                        output_file << r << g << b;
                    }

                }

                // Close file stream.
                output_file.close();
                std::cout << "Image saved." << std::endl;

            } catch (const char* err) {
                fprintf(stderr, "%s\n", err);
                output_file.close();
            }
        }

        //endregion Public Methods.

};


// Preset single colors. Mostly for testing purposes.
const Image::Rgb Image::pixelBlack = Image::Rgb(0);
const Image::Rgb Image::pixelWhite = Image::Rgb(1);
const Image::Rgb Image::pixelRed = Image::Rgb(1,0,0);
const Image::Rgb Image::pixelGreen = Image::Rgb(0,1,0);
const Image::Rgb Image::pixelBlue = Image::Rgb(0,0,1);
const Image::Rgb Image::pixelYellow = Image::Rgb(1,1,0);
const Image::Rgb Image::pixelPink = Image::Rgb(1,0,1);
const Image::Rgb Image::pixelLiBlue = Image::Rgb(0,1,1);


/*
 * Program's main. Initializes and runs program.
 */
int main(int argc, char* argv[]) {
    printf("Starting program.\n");
    printf("\n");
    printf("\n");

    // Check provided number of args.
    if (argc < 2) {
        // No args provided.
        printf("Provided too few program args. Expected one arg.");
    } else if (argc > 2) {
        // More than one arg provided.
        printf("Provided too many program args. Expected one arg.");
    } else {
        // Correct number of args provided. Pass to image class for processing.

        // Check data type.
        std::string path_str = argv[1];
        path_validator_struct* return_struct = validate_path(path_str);
        int path_type = return_struct->file_type;
        if (return_struct->err_code) {
            std::cerr << "Error verifying file type: " << return_struct->err_code.message() << std::endl;
        }
        free_path_validator_struct(return_struct);

        // Handle based on provided type.
        if (path_type == TYPE_DIR) {
            // Process dir.
            process_dir(path_str);
        } else if (path_type == TYPE_FILE) {
            // Process file.
            process_file(path_str);
        }
    }

    printf("\n");
    printf("\n");
    printf("Terminating program.\n");
    exit(0);
}


/**
 * Validates provided path and if valid, determines path type.
 * Returns struct of data.
 */
path_validator_struct* validate_path(std::string path_str) {
    int type_value = TYPE_OTHER;
    std::error_code ec;

    std::cout << "path_str: \"" + path_str + "\"\n";

    // Check if provided path exists.
    std::ifstream test_path_exists(path_str);
    if (test_path_exists) {
        // Path exists. Check if dir or file.
        if (std::filesystem::is_directory(path_str, ec)) {
            // Is dir.
            type_value = TYPE_DIR;
        } else if ((! ec) && (std::filesystem::is_regular_file(path_str, ec))) {
            // Is file.
            type_value = TYPE_FILE;
        }
    }

    // Return values.
    path_validator_struct* return_struct_ptr = initialize_path_validator_struct(type_value, ec);
    return return_struct_ptr;
}


/**
 * Finds and processes all files in directory.
 * Assumes directory is full of image files.
 */
void process_dir(std::string path_str) {

    // Process all files in directory.
    for (const auto & entry : std::filesystem::directory_iterator(path_str)) {
        // Get as string.
        std::string file_path_str{entry.path().u8string()};
        // std::cout << file_path_str << std::endl;

        // Send individual file path for processing.
        process_file(entry.path());
    }
}


/**
 * Processes a single image file.
 */
void process_file(std::string path_str) {

    Image image;
    image.display_properties();
    image.save();

    // // Open input file.
    // std::ifstream input_file;
    // input_file.open(path_str, std::ios::binary | std::ios::in);
    // if (! input_file.good()) {
    //     std::cout << "Failed to open file \"" + path_str + "\".\n";
    // }

    // // Calculate output path location.
    // std::string base_filename = path_str.substr(path_str.find_last_of("/\\") + 1);
    // std::string output_path_str = "../output/" + base_filename;

    // // Open output file.
    // std::ofstream output_file;
    // output_file.open(output_path_str, std::ios::binary | std::ios::out);
    // if (! output_file.good()) {
    //     std::cout << "Failed to open file \"" + output_path_str + "\".\n";
    // }

    // // Copy file over.
    // if (input_file.good() && output_file.good()) {
    //     // output_file << input_file;

    //     // int buffer[100];
    //     // while (input_file.read( (char*) &buffer, sizeof(buffer)) ) {
    //     //     output_file.write( (char*) &buffer, sizeof(buffer) );
    //     // }

    //     // std::copy(
    //     //     std::istreambuf_iterator<char>(input_file),
    //     //     std::istreambuf_iterator<char>( ),
    //     //     std::ostreambuf_iterator<char>(output_file)
    //     // );

    //     std::string header;
    //     input_file >> header;
    //     if (strcmp(header.c_str(), "P6") != 0) {
    //         throw("Can't read input file");
    //     } else {
    //         printf("header: %s\n\n", header.c_str());
    //     }

    //     input_file >> width >> height >> bit_depth;
    //     printf("width: %i\n", width);
    //     printf("height: %i\n", height);
    //     printf("bit depth: %i\n", bit_depth);

    //     input_file.close();
    //     output_file.close();
    // }
}



// /*
//  * Program's main. Initializes and runs program.
//  */
// int main(int argc, char* argv[]) {
//     printf("Starting program.\n");
//     printf("\n");

//     // Check provided number of args.
//     if (argc < 3) {
//         // Less than two args provided.
//         printf("Provided too few program args. Expected two integers to add.");
//     } else if (argc > 3) {
//         // More than two args provided.
//         printf("Provided too many program args. Expected two integers to add.");
//     } else {
//         // Two args provided. Compute with GPU.
//         int a = atoi(argv[1]);
//         int b = atoi(argv[2]);
//         printf("Adding %i and %i\n", a, b);
//         compute(a, b);
//     }

//     printf("\n");
//     printf("Terminating program.\n");
//     exit(0);
// }


// /*
//  * Function to have CPU and GPU collaborate for basic addition computation.
//  */
// void compute(int a, int b) {
//     int sum;
//     int* dev_sum;

//     // Allocate memory on the GPU.
//     cudaMalloc((int**) &dev_sum, sizeof(int));

//     // Call CUDA function.
//     cuda_add<<<1,1>>>(a, b, dev_sum);

//     // Copy GPU result back to CPU.
//     cudaMemcpy(&sum, dev_sum, sizeof(int), cudaMemcpyDeviceToHost);

//     // Display result.
//     printf("Sum is %i\n", sum);

//     cudaFree(dev_sum);
// }


// /**
//  * Entrypoint to GPU kernel execution.
//  */
// __global__ void cuda_add(int a, int b, int* c) {
//     *c = a + b;
// }
